from ._internal import DEFAULT, defaultarguments, DoreahConfig
import math
import string
import re
import os
import yaml, json
from jinja2 import Environment, PackageLoader, select_autoescape

## for these configuration objects:
## 		NONE always is equivalent to the setting not being specified - i.e. fallback to default
##		FALSE can be used as a sentinel regardless of type
##
##		parsing and conversion of values should be done client-side. the server only checks if valid and rejects if not




config = DoreahConfig("configuration",
)


JINJAENV = Environment(
	loader=PackageLoader('doreah', 'resources/configuration'),
	autoescape=select_autoescape(['html', 'xml'])
)


class Configuration:


	def __init__(self,settings,configfile="settings.ini",save_endpoint="/settings",env_prefix=None):
		flatten = {key.lower():settings[category][key] for category in settings for key in settings[category]}
		self.categories = {cat:[k for k in settings[cat]] for cat in settings}
		self.types = {k:flatten[k][0] for k in flatten}
		self.names = {k:flatten[k][1] for k in flatten}
		self.descs = {k:flatten[k][3] for k in flatten if len(flatten[k]) > 3}

		self.usersettings = {}
		self.environment = {}
		self.defaults = {k:flatten[k][2] for k in flatten}
		#self.expire = {k:flatten[k][3] if len(flatten[k])>3 else -1 for k in flatten}

		self.configfile = configfile
		self.save_endpoint = save_endpoint
		self.env_prefix = env_prefix

		self.load_environment()
		self.load_from_file()

	def update(self,settings):
		for k in settings:
			if settings[k] is None: del self.usersettings[k]
			else: self.usersettings[k] = self.types[k].sanitize(settings[k])
		self.write_to_file()

	def get_setting_info(self,key):
		return {
			"key":key.lower(),
			"value":self.get_active(key),
			"default":self.get_default(key),
			"name":self.names[key.lower()],
			"description":self.descs.get(key.lower()) or None
		}


	def get_user(self,key):
		return self.usersettings.get(key.lower())
	def get_env(self,key):
		return self.environment.get(key.lower())
	def get_default(self,key):
		return self.defaults.get(key.lower())


	# this method gets the setting only if it's not the web/file-adjustable one
	def get_fallback(self,key):
		for layer in (self.get_env,self.get_default):
			val = layer(key)
			if val is not None: return val
		return None

	# this method gets the setting only if it has in any way been user-specified
	def get_specified(self,key):
		for layer in (self.get_user,self.get_env):
			val = layer(key)
			if val is not None: return val
		return None

	# get the one that's valid
	def get_active(self,key):
		v = self.get_user(key)
		if v is None: v = self.get_fallback(key)
		return v

	### MAGIC

	def __getitem__(self,key):
		# special case to support easier transition from settings module
		if isinstance(key,tuple):
			return (self[k] for k in key)
		return self.get_active(key)
		#if key in self.usersettings: return self.usersettings[key]
		#if key in self.defaults: return self.defaults[key]
		#return None

	def __setitem__(self,key,value):
		self.usersettings[key.lower()] = value
		self.write_to_file()

	def __iter__(self):
		return (k for k in self.defaults)

	def __len__(self):
		return len(self.usersettings)

	def todict(self):
		d = {k:self.defaults[k] for k in self.defaults}
		d.update(self.usersettings)
		return d

	def html(self):
		template = JINJAENV.get_template("settings.jinja")
		return template.render({"configuration":self})

	def load_environment(self):
		if self.env_prefix is not None:
			for k in os.environ:
				if k.startswith(self.env_prefix.upper()):
					sk = k[len(self.env_prefix):].lower()
					if sk in self.types:
						self.environment[sk] = self.types[sk].fromstring(os.environ[k])
			#self.environment = {k[len(self.env_prefix):].lower():os.environ[k] for k in os.environ if k.startswith(self.env_prefix.upper())}

	def load_from_file(self):
		ext = self.configfile.split(".")[-1].lower()
		try:
			settings = formats.loaders[ext].load_file(self.configfile)
			settings = {k.lower():settings[k] for k in settings}
			settings = {
				k:settings[k] for k in settings if
				k in self.types and (self.types[k].validate(settings[k]) or settings[k] is False)
			}
			self.usersettings = settings
		except FileNotFoundError:
			self.write_to_file()

	def write_to_file(self):
		ext = self.configfile.split(".")[-1].lower()
		try:
			formats.loaders[ext].write_file(self.configfile,self.usersettings)
		except:
			print("Could not write file",self.configfile)
			raise


	def render_help(self,targetfile,top_text=""):
		template = JINJAENV.get_template("settings.md.jinja")
		txt = template.render({"configuration":self,"top_text":top_text})
		with open(targetfile,"w") as fd:
			fd.write(txt)

# this is just a namespace, not a real class
class types:

	class SettingType:
		def __init__(self):
			pass

		def __desc__(self):
			return "Generic Datatype"

		# render the html element for the web interface
		def html(self,active,default,user,setting):
			template = JINJAENV.get_template(self.jinjafile())
			return template.render({"type":self,"active":active,"default":default,"user":user,"setting":setting})

		def jinjafile(self):
			return "types/" + self.__class__.__name__ + ".jinja"

		# block invalid values
		def sanitize(self,input):
			if self.validate(input) or input is False: # False is always allowed
				return input
			else:
				return self.default

		# return how it should be displayed in an html settings field
		def html_value(self,value):
			return value

		# parse string inputs
		def fromstring(self,input):
			return input

	class Integer(SettingType):
		regex = "[0-9]+"
		default = 0

		def __init__(self,min=-math.inf,max=math.inf):
			self.min = min
			self.max = max

		def __desc__(self):
			return "Integer"

		def validate(self,input):
			return isinstance(input,int) and input >= self.min and input <= self.max

		def fromstring(self,input):
			try: return int(input)
			except: return None

	class String(SettingType):
		regex = "([^'\"]*)"
		default = ""

		def __init__(self,minlength=0,maxlength=math.inf):
			self.minlength = minlength
			self.maxlength = maxlength

		def __desc__(self):
			return "String"

		def validate(self,input):
			return isinstance(input,str) and len(input) >= self.minlength and len(input) <= self.maxlength

		def html_value(self,value):
			if value in [None,False]: return ""
			else: return value

	class ASCIIString(String):
		regex = r"\w*"

		def __desc__(self):
			return "ASCII String"

		def validate(self,input):
			return isinstance(input,str) and len(input) >= self.minlength and len(input) <= self.maxlength

	class Choice(SettingType):
		def __init__(self,options=()):
			self.options = options

		def __desc__(self):
			return "Choice"

		def validate(self,input):
			return input in self.options

	class MultiChoice(Choice):

		def __desc__(self):
			return "Multiple Choice"
		def validate(self,input):
			return all([i in self.options for i in input])

	class Set(SettingType):
		default = []

		def __init__(self,type,minmembers=0,maxmembers=math.inf):
			self.type = type
			self.minmembers = minmembers
			self.maxmembers = maxmembers

		def __desc__(self):
			return "Set"

		def validate(self,input):
			return len(input) <= self.maxmembers and len(input) >= self.minmembers and all(self.type.validate(m) for m in input)

	class List(SettingType):
		default = []

		def __init__(self,type,minmembers=0,maxmembers=math.inf):
			self.type = type
			self.minmembers = minmembers
			self.maxmembers = maxmembers

		def __desc__(self):
			return "List"

		def validate(self,input):
			return len(input) <= self.maxmembers and len(input) >= self.minmembers and all(self.type.validate(m) for m in input)

	class Boolean(SettingType):
		default = False

		def __desc__(self):
			return "Boolean"

		def validate(self,input):
			return input in [True,False]

		def fromstring(self,input):
			if input.lower() in ['true','yes','y','on']: return True
			if input.lower() in ['false','no','n','off']: return False
			return None


class formats:
	class GenericHandler:

		def load_file(self,filename):
			with open(filename) as descriptor:
				return self.load_descriptor(descriptor)

		def write_file(self,filename,data):
			with open(filename,"w") as descriptor:
				return self.write_descriptor(descriptor,data)


		def load_descriptor(self,descriptor):
			txt = descriptor.read()
			return self.text_to_data(txt)

		def write_descriptor(self,descriptor,data):
			txt = self.data_to_text(data)
			return descriptor.write(txt)


	class YAMLHandler(GenericHandler):

		def load_descriptor(self,descriptor):
			return yaml.safe_load(descriptor)

		def write_descriptor(self,descriptor,data):
			return yaml.dump(data,descriptor)

	class JSONHandler(GenericHandler):

		def load_descriptor(self,descriptor):
			return json.load(descriptor)

	class INIHandler(GenericHandler):

		boolx = lambda x: True if x.lower() in ['true','yes','y'] else False if x.lower() in ['false','no','n'] else None
		regex_key = 					re.compile(r"([a-zA-Z][\w]+)"	) # at least 2 characters
		regex_values = [
			("singlequoted",	str,	re.compile(r"'([^']*)'")),
			("doublequoted",	str,	re.compile(r'"([^"]*)"')),
			("boolean",			boolx,	re.compile(r"(false|no|n|true|yes|y)",re.IGNORECASE)),
			("identifier",		str,	re.compile(r"([a-zA-Z][\w]+)")),
			("integer",			int,	re.compile(r"([-+]?[\d]+)")),
			("float",			float,	re.compile(r"([\d]+[.]?[\d]*)"))
		]
		regex_separator = 				re.compile(r"[ \t]*=[ \t]*")

		regex_value = re.compile("(" + "|".join(exp.pattern for nm,ty,exp in regex_values) + ")")
		regex_fullline = re.compile("".join(exp.pattern for exp in [
			regex_key,
			regex_separator,
			regex_value
		]))


		def text_to_data(self,txt):
			result = {}
			for l in txt.split("\n"):
				match = self.regex_fullline.fullmatch(l)
				if match:
					key, val, *_ = match.groups()
					for nm,ty,exp in self.regex_values:
						match = exp.fullmatch(val)
						if match:
							val = match.groups()[0]
							result[key] = ty(val)
							break

			return result

		def data_to_text(self,data):
			lines = []
			for k in data:
				if isinstance(data[k],str):
					if '"' not in data[k]: val = f'"{data[k]}"'
					elif "'" not in data[k]: val = f"'{data[k]}'"
				else: val = str(data[k])
				lines.append(k.upper() + " = " + val)
			lines.append("")

			return "\n".join(lines)


	loaders = {
		"yaml":YAMLHandler(),
		"yml":YAMLHandler(), #yes, two instances for no reason other than cleaner code
		"json":JSONHandler(),
		"ini":INIHandler()
	}



from threading import Thread, Timer
import datetime

from ._internal import DEFAULT, defaultarguments, gopen, DoreahConfig

_yearly_funcs = []
_monthly_funcs = []
_daily_funcs = []


config = DoreahConfig("regular",
	autostart=True,
	offset=0
)

def tz():
	return datetime.timezone(offset=datetime.timedelta(hours=config["offset"]))


def yearly(func):
	"""Decorator for yearly functions. Depending on configuration, is equivalent to either :meth:`runyearly` or :meth:`repeatyearly`."""
	if config["autostart"]: return runyearly(func)
	else: return repeatyearly(func)

def monthly(func):
	"""Decorator for monthly functions. Depending on configuration, is equivalent to either :meth:`runmonthly` or :meth:`repeatmonthly`."""
	if config["autostart"]: return runmonthly(func)
	else: return repeatmonthly(func)

def daily(func):
	"""Decorator for daily functions. Depending on configuration, is equivalent to either :meth:`rundaily` or :meth:`repeatdaily`."""
	if config["autostart"]: return rundaily(func)
	else: return repeatdaily(func)

def hourly(func):
	"""Decorator for hourly functions. Depending on configuration, is equivalent to either :meth:`runhourly` or :meth:`repeathourly`."""
	if config["autostart"]: return runhourly(func)
	else: return repeathourly(func)




def runyearly(func):
	"""Decorator to make the function execute on first definition as well as every year."""

	def self_scheduling_func():
		# execute function
		func()

		# schedule next execution
		now = datetime.datetime.now(tz=tz())
		nextyear = datetime.datetime(now.year+1,1,1,tzinfo=tz())
		wait = nextyear.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func).start()

	# now execute it for the first time
	t = Timer(5,self_scheduling_func)
	t.daemon = True
	t.start()
	return func


def repeatyearly(func):
	"""Decorator to make the function repeat every new year after being called once."""

	def self_scheduling_func(*args,**kwargs):
		# execute function
		func(*args,**kwargs)

		# schedule next execution
		now = datetime.datetime.now(tz=tz())
		nextyear = datetime.datetime(now.year+1,1,1,tzinfo=tz())
		wait = nextyear.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func,args=args,kwargs=kwargs).start()

	def starter(*args,**kwargs):
		t = Thread(target=self_scheduling_func,args=args,kwargs=kwargs)
		t.daemon = True
		t.start()
	return starter

def runmonthly(func):
	"""Decorator to make the function execute on first definition as well as every month."""

	def self_scheduling_func():
		# execute function
		func()

		# schedule next execution
		now = datetime.datetime.now(tz=tz())
		nextmonth = datetime.datetime(now.year,now.month + 1,1,tzinfo=tz()) if now.month != 12 else datetime.datetime(now.year+1,1,1,tzinfo=tz())
		wait = nextyear.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func).start()

	# now execute it for the first time
	t = Timer(5,self_scheduling_func)
	t.daemon = True
	t.start()
	return func


def repeatmonthly(func):
	"""Decorator to make the function repeat every new month after being called once."""

	def self_scheduling_func(*args,**kwargs):
		# execute function
		func(*args,**kwargs)

		# schedule next execution
		now = datetime.datetime.now(tz=tz())
		nextmonth = datetime.datetime(now.year,now.month + 1,1,tzinfo=tz()) if now.month != 12 else datetime.datetime(now.year+1,1,1,tzinfo=tz())
		wait = nextyear.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func,args=args,kwargs=kwargs).start()

	def starter(*args,**kwargs):
		t = Thread(target=self_scheduling_func,args=args,kwargs=kwargs)
		t.daemon = True
		t.start()
	return starter

def rundaily(func):
	"""Decorator to make the function execute on first definition as well as every day."""

	def self_scheduling_func():
		# execute function
		func()

		# schedule next execution
		now = datetime.datetime.now(tz=tz())
		nextday = datetime.datetime(now.year,now.month,now.day,tzinfo=tz()) + datetime.timedelta(days=1)
		wait = nextday.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func).start()

	# now execute it for the first time
	t = Timer(5,self_scheduling_func)
	t.daemon = True
	t.start()
	return func


def repeatdaily(func):
	"""Decorator to make the function repeat every new day after being called once."""

	def self_scheduling_func(*args,**kwargs):
		# execute function
		func(*args,**kwargs)

		# schedule next execution
		now = datetime.datetime.now(tz=tz())
		nextday = datetime.datetime(now.year,now.month,now.day,tzinfo=tz()) + datetime.timedelta(days=1)
		wait = nextday.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func,args=args,kwargs=kwargs).start()

	def starter(*args,**kwargs):
		t = Thread(target=self_scheduling_func,args=args,kwargs=kwargs)
		t.daemon = True
		t.start()
	return starter



def runhourly(func):
	"""Decorator to make the function execute on first definition as well as every hour."""

	def self_scheduling_func():
		# execute function
		func()

		# schedule next execution
		now = datetime.datetime.utcnow()
		nexthour = datetime.datetime(now.year,now.month,now.day,now.hour) + datetime.timedelta(hours=1)
		wait = nexthour.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func).start()

	# now execute it for the first time
	t = Timer(5,self_scheduling_func)
	t.daemon = True
	t.start()
	return func


def repeathourly(func):
	"""Decorator to make the function repeat every new hour after being called once."""

	def self_scheduling_func(*args,**kwargs):
		# execute function
		func(*args,**kwargs)

		# schedule next execution
		now = datetime.datetime.utcnow()
		nexthour = datetime.datetime(now.year,now.month,now.day,now.hour) + datetime.timedelta(hours=1)
		wait = nexthour.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func,args=args,kwargs=kwargs).start()

	def starter(*args,**kwargs):
		t = Thread(target=self_scheduling_func,args=args,kwargs=kwargs)
		t.daemon = True
		t.start()
	return starter





#for testing
def _runoften(func):
	def self_scheduling_func():
		# execute function
		func()

		# schedule next execution
		wait = 15
		Timer(wait,self_scheduling_func).start()


	# now execute it for the first time

	t = Timer(5,self_scheduling_func)
	t.daemon = True
	t.start()
	return func


def _repeatoften(func):
	def self_scheduling_func(*args,**kwargs):
		# execute function
		func(*args,**kwargs)

		# schedule next execution
		wait = 15
		Timer(wait,self_scheduling_func,args=args,kwargs=kwargs).start()

	def starter(*args,**kwargs):
		t = Thread(target=self_scheduling_func,args=args,kwargs=kwargs)
		t.daemon = True
		t.start()
	return starter



import pickle
import yaml
import os


from ._internal import DEFAULT, defaultarguments, DoreahConfig

config = DoreahConfig("database")



class Ref:
	def __init__(self,cls,backref=None,exclusive=False):
		self.cls = cls
		self.backref = backref
		self.exclusive = exclusive

class MultiRef(Ref):
	pass

class Primary:
	pass

class Database:

	def _getclassbyname(self,cls):
		return [c for c in self.class_to_objects if c.__name__ == cls][0]

	def getall(self,cls):
		if isinstance(cls,str): cls = self._getclassbyname(cls)
		return self.class_to_objects.get(cls)[:]

	def get(self,id):
		return self.id_to_object[id]

	def getby(self,cls,**keys):
		if isinstance(cls,str): cls = self._getclassbyname(cls)
		tup = tuple(tuplify(keys[k],ignore_capitalization=cls.__dbsettings__.get("ignore_capitalization")) for k in cls.__primarykey__)
		return self.class_primary_keys[cls][tup]

	def getby_or_create(self,cls,**keys):
		if isinstance(cls,str): cls = self._getclassbyname(cls)

		try:
			return self.getby(cls,**keys)
		except:
			return cls(**keys)

	def delete(self,inp):
		# can take uid or reference
		if isinstance(inp,int):
			uid = inp
			obj = self.id_to_object[uid]
		else:
			uid = inp.uid
			obj = inp

		cls = obj.__class__

		self.id_to_object[uid] = None
		self.class_to_objects[cls].remove(obj)
		primkey = tuple(tuplify(obj.__getattribute__(v),ignore_capitalization=cls.__dbsettings__.get("ignore_capitalization")) for v in cls.__primarykey__)
		del self.class_primary_keys[cls][primkey]

	def __init__(self_db,file="database.ddb"):
		#print("Initializing database...")
		self_db.file = file

		self_db.counter = 0


		self_db.id_to_object = {}
		self_db.class_to_objects = {}
		self_db.class_primary_keys = {}

		self_db.loadarea = {}
		self_db.load()

		# every instance creates a class to inherit from
		class obj:
			__dbsettings__ = {"ignore_capitalization":False}
			def __repr__(self):
				try:
					name = self.__db_repr__()
				except:
					for attr in ["name","title","identifier"]:
						try:
							name = getattr(self,attr)
							break
						except:
							pass
					else:
						name = "Unknown"
				return "<[" + str(self.uid) + "] " + self.__class__.__name__  + " '" + name + "'>"

			def __init_subclass__(cls):
				# register class
				self_db.class_to_objects[cls] = []
				self_db.class_primary_keys[cls] = {}
				cls.__primarykey__ = tuple()
				#print("New database class defined:",cls)


				types = cls.__annotations__


				# add to referenced classes
				for v in types:
					if v in cls.__dict__:
						classvar = cls.__dict__[v]
						if isinstance(classvar,Ref) and classvar.backref is not None:
							# make getter method that checks instances of THIS object for references to the target object


							# 1 to 1
							if classvar.exclusive and not isinstance(classvar,MultiRef):
								def find_object_that_references_me(self,attr=v):
									for obj in self_db.class_to_objects[cls]:
										if obj.__getattribute__(attr) is self: return obj
								prop = property(find_object_that_references_me)

							# many to 1
							elif not classvar.exclusive and not isinstance(classvar,MultiRef):
								def find_objects_that_reference_me(self,attr=v):
									l = []
									for obj in self_db.class_to_objects[cls]:
										if obj.__getattribute__(attr) is self: l.append(obj)
									return l
								prop = property(find_objects_that_reference_me)

							# 1 to many
							elif classvar.exclusive and isinstance(classvar,MultiRef):
								def find_object_that_references_me_among_others(self,attr=v):
									for obj in self_db.class_to_objects[cls]:
										if self in obj.__getattribute__(attr): return obj
								prop = property(find_object_that_references_me_among_others)

							# many to many
							elif not classvar.exclusive and isinstance(classvar,MultiRef):
								def find_objects_that_reference_me_among_others(self,attr=v):
									l = []
									for obj in self_db.class_to_objects[cls]:
										if self in obj.__getattribute__(attr): l.append(obj)
									return l
								prop = property(find_objects_that_reference_me_among_others)


							setattr(classvar.cls,classvar.backref,prop)

					#	# save what defines our primary key (in addition to uid)
					#	elif isinstance(classvar,Primary):
					#		cls.__primarykey__.append(v)

				#cls.__primarykey__ = tuple(cls.__primarykey__)
				if hasattr(cls,"__primary__"): cls.__primarykey__ = tuple(cls.__primary__)


				# create constructor

				def init(self,types=types,force_uid=None,**vars):
					alreadyexisting = hasattr(self,"uid") # if we have an existing class, don't do all of the init
					for v in vars:
						if vars[v] is None: continue
						# set attributes according to keyword arguments
						try:
							assert isinstance(vars[v],types[v])
						except:
							print("Wrong data type!",v,"should be",types[v],"not",type(vars[v]))
							raise
						setattr(self,v,vars[v])



					if not alreadyexisting:
						# set defaults
						for v in types:
							if v not in vars or vars[v] is None:
								if types[v] in [str,int,float,list]:
									setattr(self,v,types[v]()) # can just call type to get a version of it, e.g. list() -> []
								else:
									# if we have a custom type (another db type), we might not be able
									# to call it without primary key. we don't want to generate a fake one
									# that is saved in the database just because we don't have this attribute yet
									setattr(self,v,None)


						# register object with Database
						self.uid = force_uid if force_uid is not None else self_db.counter
							#max(self_db.id_to_object) + 1 if len(self_db.id_to_object) > 0 else 0

						self_db.id_to_object[self.uid] = self
						self_db.class_to_objects[cls].append(self)
						while self_db.counter in self_db.id_to_object: self_db.counter += 1

						if len(cls.__primarykey__) > 0:
							primkey = tuple(tuplify(vars[v],ignore_capitalization=cls.__dbsettings__.get("ignore_capitalization")) for v in cls.__primarykey__)
							self_db.class_primary_keys[cls][primkey] = self

				cls.__init__ = init


				# create creator (if primary keys match, no new object is created!)
				# we cant programmatically assign a metaclass, so we can't avoid __init__
				# being called
				def new(cls,**kwargs):

					try:
						return self_db.getby(cls,**kwargs)
					except Exception as e:
						inst = object.__new__(cls)
						return inst

				cls.__new__ = new



				# after each class is defined, check if we can create new objects
				self_db.inject()

		self_db.DBObject = obj


	def save(self_db):

		d = {}
		for cls in self_db.class_to_objects:
			d[cls.__name__] = []
			for obj in self_db.class_to_objects[cls]:
				objd = {"uid":obj.uid}
				for var in cls.__annotations__:
					objd[var] = yamlify(getattr(obj,var))
				d[cls.__name__].append(objd)

		tmpfile = self_db.file + ".tmp"
		folder = os.path.dirname(tmpfile)
		os.makedirs(folder,exist_ok=True)
		with open(tmpfile,"w") as f:
			yaml.dump(d,f)
		os.replace(tmpfile,self_db.file)

		#print("Database saved to disk.")


	def load(self_db):
		try:
			with open(self_db.file,"r") as f:
				self_db.loadarea = yaml.safe_load(f)
				# temporary storage until proper classes are defined
			#print("Database loaded into memory.")
			self_db.counter = max([0] + [obj["uid"] for cls in self_db.loadarea for obj in self_db.loadarea[cls]])
			self_db.inject()
		except:
			#print("No existing database found.")
			pass




	def inject(self_db):
		classnames = [c for c in self_db.loadarea]

		# go through classes twice in case later classes enable references in earlier ones
		# twice is enough because we run this after every new class definition
		for clsname in classnames + list(reversed(classnames)):
			# check if class of this name exists
			classes = [c for c in self_db.class_to_objects if c.__name__ == clsname]
			if len(classes) > 0:
				cls = classes[0]

				for instance in self_db.loadarea[clsname]:


					for attr in cls.__annotations__:

						if hasattr(cls,attr) and isinstance(getattr(cls,attr),MultiRef) and all(isinstance(el,int) for el in instance[attr]):
							# convert uid to actual reference
							try:
								uids = instance[attr]
								instance[attr] = [self_db.get(uid) for uid in uids]
							except KeyError as exc:
								break
						elif hasattr(cls,attr) and isinstance(getattr(cls,attr),Ref) and isinstance(instance[attr],int):
							# convert uid to actual reference
							try:
								uid = instance[attr]
								instance[attr] = self_db.get(uid)
							except KeyError as exc:
								break
					else:
						# no break, so all references have been converted, we can use this object!
						uid = instance.pop("uid")
						cls(**instance,force_uid=uid)

						# delete from load area
						self_db.loadarea[clsname] = [i for i in self_db.loadarea[clsname] if i is not instance]






def yamlify(obj):
	if type(obj) in [str,int,float,bool]: return obj
	if obj == [] or obj == {}: return obj
	if obj is None: return None

	try:
		return {k:yamlify(obj[k]) for k in obj}
	except:
		pass

	try:
		return [yamlify(e) for e in obj]
	except:
		pass

	return obj.uid


def tuplify(obj,ignore_capitalization=False):
	if type(obj) is str and ignore_capitalization: return obj.lower()
	if type(obj) in [str,int,float]: return obj
	if obj is None: return None

	try:
		return frozenset(tuplify(e) for e in obj)
	except:
		pass

	return obj.uid



from ._internal import DEFAULT, defaultarguments, DoreahConfig


config = DoreahConfig("datatypes",
)

class DictStack:

	def __init__(self,*args,**kwargs):
		self.stack = [dict(*args,**kwargs)]
		#same options as normal dict, either kwargs for entries or one arg for dict

	def __flat__(self):
		di = {}
		for d in self.stack:
			di.update(d)
		return di

	def __setitem__(self,key,value):
		for d in reversed(self.stack):
			if key in d:
				d[key] = value
				break
		else:
			self.stack[0][key] = value

	def __getitem__(self,key):
		for d in reversed(self.stack):
			if key in d:
				return d[key]
		raise KeyError

	def __contains__(self,key):
		return self.__flat__().__contains__(key)

	def __delitem__(self,key):
		for d in reversed(self.stack):
			if key in d:
				del d[key]
		raise KeyError

	def clear(self):
		self.stack = [{}]

	def copy(self):
		return [dict(d) for d in self.stack]

	def get(self,key):
		try:
			return self.__getitem__(key)
		except:
			return None

	def has_key(self,key):
		return self.__contains__(key)

	def update(self,*args,**kwargs):
		self.stack[-1].update(*args,**kwargs)

	def __repr__(self):
		return self.__flat__().__repr__()

	def __len__(self):
		return self.__flat__().__len__()

	def keys(self):
		return self.__flat__().keys()
	def values(self):
		return self.__flat__().values()
	def items(self):
		return self.__flat__().items()

	def push(self,*args,**kwargs):
		self.stack.append(dict(*args,**kwargs))
	def pop(self):
		return self.stack.pop()

	@property
	def locl(self):
		return self.stack[-1]
	@property
	def globl(self):
		return self.stack[0]

	@property
	def height(self):
		return len(self.stack)


	def newlayer(self,*args,**kwargs):
		return _DictStackContext(stack=self,height=self.height,dict=dict(*args,**kwargs))

class _DictStackContext:

	def __init__(self,stack,height,dict):
		self.stack = stack
		self.height = height
		self.dict = dict

	def __enter__(self):
		assert self.stack.height == self.height
		self.stack.push(self.dict)

	def __exit__(self,*exc):
		assert self.stack.height == self.height + 1
		self.stack.pop()




__all__ = [
	"auth",
	"caching",
	"control",
	"database",
	"datatypes",
	"filesystem",
	"fs",
	"io",
	"logging",
	"packageutils",
	"persistence",
	"pyhp",
	"regular",
	"scraping"
	"settings",
	"timing",
	"tsv"

]

from ._internal import config


# useful things for everyone


def deprecated(func):
	"""Function decorator to deprecate a function"""

	def newfunc(*args,**kwargs):
		print("\033[93m" + "Function " + func.__name__ + " is deprecated!" + "\033[0m")
		return func(*args,**kwargs)

	return newfunc
